#include "operator.hh"

void Operator::Modulate(Operator &o)
{
	o.chain     = this;
	o.chaintype = mod_freq;
}

void Operator::AddTo(Operator &o)
{
	o.chain     = this;
	o.chaintype = mod_add;
}

void Operator::Unlink(void)
{
	chain = 0;
}

const unsigned Operator::multiplyfreq(unsigned freq) const
{
	switch(freqmul)
	{
		case 0: return freq/2;
		case 1: return freq;
		case 11: return freq * 10;
		case 13: return freq * 12;
		case 14: return freq * 15;
	}
	return freq * freqmul;
}

Operator &OPL2_link(
    Operator &op1, Operator &op2,
    unsigned mode)
{
	op1.Unlink();
	op2.Unlink();
	
	switch(mode)
	{
		case 0: op1.Modulate(op2); break;
		case 1: op1.AddTo(op2);    break;
	}
	return op2;
}

Operator &OPL3_link(
    Operator &op1, Operator &op2,
    Operator &op3, Operator &op4,
    unsigned mode)
{
	op1.Unlink();
	op2.Unlink();
	op3.Unlink();
	op4.Unlink();
	
	switch(mode)
	{
		case 0: op1.Modulate(op2); 
		        op2.Modulate(op3);
		        op3.Modulate(op4);
		        return op4;

		case 1: break;
		
		case 2: op2.Modulate(op3);
		        op3.Modulate(op4);
		        op4.AddTo(op1);
		        return op1;
		
		case 3: op2.Modulate(op3);
		        op3.AddTo(op1);
		        op1.AddTo(op4);
		        return op4;
	}
	
	/* Compatibility mode */
	op1.Modulate(op2);
	op3.Modulate(op4);
	op2.AddTo(op4);
	return op4;
}

void Operator::Write20(unsigned char value)
{
	tremolo = value & 0x80;
	vibrato = value & 0x40;
	dosustain=value & 0x20;
	KSR =     value & 0x10;
	freqmul = value & 15;
}

void Operator::Write40(unsigned char value)
{
	KSL         = value >> 6;
	outputlevel = value & 63;
}

void Operator::Write60(unsigned char value)
{
	attack   = value >> 4;
	decay    = value & 15;
}
	
void Operator::Write80(unsigned char value)
{
	sustain  = value >> 4;
	release  = value & 15;
}

unsigned OperatorPortOffset(unsigned channel)
{
	unsigned ofs = 0x100 * (channel / 18);
	register unsigned chan18 = channel % 18;
	return ofs + (chan18%6) + 2 * (chan18/6);
}

unsigned PortOffsetOperator(unsigned offset)
{
	unsigned chan = 18 * (offset / 0x100);
	register unsigned low = offset & 0x1F;
	
	chan = low;
	
	if(low >= 0x08)chan -= 2;
	if(low >= 0x10)chan -= 2;
	
	if(chan >= 36)chan = 36-1;
	
	return chan;
}
